/*
 * Progress.cpp
 *
 *  Created on: Aug 15, 2011
 *      Author: janez
 */

#include "Progress.h"

#include <fstream>
#include <iostream>
#include <string>

using namespace std;


Progress::Progress() {
	mark();
}

Progress::~Progress() {

}

/**
 * Initialize indicator at given position. Do nothing if position is out of limits.
 */
void Progress::init( unsigned int pos ){
	if( pos < 0 || pos > indicators.size()-1 ) {
		// position out of limits;
		return;
	}

	// initialize progress structure at given position
	indicators[pos].mark = 0;
	indicators[pos].format = "";
	indicators[pos].stepNextPrint = 0;
	indicators[pos].stepDelta = 0;
	indicators[pos].dump = false;
	indicators[pos].toFile = "";
}


/**
 * Calculate difference between times.
 */
struct timeval Progress::time_diff( struct timeval t2, struct timeval t1 ) {
	struct timeval diff, t3;

	if( t2.tv_sec < t1.tv_sec ) {
		// second time is bigger; swap the times
		t3 = t2;
		t2 = t1;
		t1 = t3;
	}

	// now calculate...
	diff.tv_sec = t2.tv_sec - t1.tv_sec;

	if( t2.tv_usec < t1.tv_usec ) {
		// subtraction + correct seconds
		diff.tv_usec = 1000 + t2.tv_usec - t1.tv_usec;
		diff.tv_sec --;
	} else {
		// normal subtraction
		diff.tv_usec = t2.tv_usec - t1.tv_usec;
	}

	return diff;
}


/**
 * Convert difference between two times into double
 */
double Progress::time_diff_d( struct timeval t2, struct timeval t1 ) {
	struct timeval t_diff;
	double d = 0;

	t_diff = time_diff(t2, t1);
	d = t_diff.tv_sec;
	d += (t_diff.tv_usec / 1000.);

	return d;
}




/**
 * Return string date in format YYYY-MM-DD HH:MM:SS.ususus for given time
 */
string Progress::timeString(struct timeval* t) {
	char tmp[200];
	string ret;
	struct tm* a;

	// get time structure for displaying
	a = gmtime(&(t->tv_sec));
	// format a string
	sprintf(tmp, "%4d-%2d-%2d %02d:%02d:%02d.", a->tm_year, a->tm_mon, a->tm_mday, a->tm_hour, a->tm_min, a->tm_sec);
	ret = tmp + t->tv_usec;

	return ret;
}


/**
 * Add values for indicator: at which % to print indicator and what to print (format may include printf format characters)
 */
void Progress::addIndicator( double inMark, std::string inFormat ) {
	struct indicator ind;

	ind.mark = inMark;
	ind.format = inFormat;
	ind.dump = false;
	// calculate number of steps between displaying indicators and
	// next step when indicator will be displayed

	if( inMark == 0 ) {
		// display every step
		ind.stepDelta = 1L;
		ind.stepNextPrint = ind.stepDelta;

	} else {
		// compute next step
		ind.stepDelta = (long int) (maxStep * inMark);
		ind.stepNextPrint = ind.stepDelta;
	}

	// add to array of indicators
	indicators.push_back(ind);
}



/**
 * Add values for indicator to write values to file (dump).
 */
void Progress::addIndicator( double inMark, std::string toFile, std::string whatToDisplay ) {
	struct indicator ind;

	ind.mark = inMark;
	ind.format = whatToDisplay;
	ind.toFile = toFile;
	ind.dump = true;
	// calculate number of steps between displaying indicators and
	// next step when indicator will be displayed
	ind.stepDelta = (long int) (maxStep * inMark);
	ind.stepNextPrint = ind.stepDelta;

	// add to array of indicators
	indicators.push_back(ind);
}



/**
 * Display given message to stdout or file, depending of outpurFilename. This method is used
 * internally for displaying.
 */
void Progress::displayMessage(std::string mess){
	if( outFile == "" ) {
		// display on stdout
		cout << mess;
		cout.flush();

	} else {
		// write progress to file with given name; append to file
		fstream fp;

		fp.open(outFile.c_str(), fstream::app | fstream::out);

		if( fp.is_open() ) {
			// OK, write message to file and close the file
			fp << mess;
			fp.close();

		} else {
			// display error to stdout
			cout << "Progress: can not open file " << outFile << " for writing progress into it." << endl;
		}
	}
}




/**
 * Display current progress.
 */
void Progress::display(long step) {
	double t_diff;
	string mess = "";
	struct timeval now, diff;

	// get current time
	gettimeofday(&now, NULL);

	// calculate difference between times
	diff = time_diff( now, tstamp);
	t_diff = time_diff_d( now, tstamp );

	// prepare string to display
	mess = "[";
	mess += timeString(&now);
	mess += "] Step: " + step;
	mess += ", (";
	mess += t_diff;
	mess += ")\n";

	// display message
	displayMessage( mess );

}



/**
 * Check all indicator and display one that we should for given step.
 */
void Progress::displayIndicator(long int step) {
	unsigned int i;
	char fmt[100];
	string mess = "";


	for( i=0 ; i < indicators.size() ; i++ ) {
		// check all defined indicators...
		if( indicators[i].stepNextPrint <= step+1 ) {
			// yes, display current indicator
			sprintf(fmt, indicators[i].format.c_str(), 100. * ((step+1) * 1.0 / maxStep) );
			mess += fmt;
			// calculate and store step number at which next indicator is displayed
			indicators[i].stepNextPrint += indicators[i].stepDelta;
		}
	}

	if( mess != "" ) {
		// we have something to display, so do it
		displayMessage( mess );
	}
}
