/*
 * Simulation.cpp
 *
 *  Created on: Aug 7, 2011
 *      Author: janez
 */

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <stdlib.h>
#include "Simulation.h"

using namespace std;

//#define DEBUG

Simulation::Simulation() {
	simStep = maxSimStep = 0;
	t = 0;
	dt = 0;
	inputFileName = inputDumpFileName = outputDumpFilename = "";
	// are following variables defined or not
	defMaxSimStep = defdt = false;
	// default is not test mode for input file
	inputFileTest = false;

	tree.init();
	particles.init();
	species.clear();
	sources.clear();
	initLoad.clear();
	potential.clear();
	fields.clear();
	diags.clear();

	// initialize the pointer to compute class
	compute = NULL;
	mover = NULL;

	numParticlesRightOut = 0;
	numParticlesLeftOut = 0;

}

Simulation::~Simulation() {

	if( compute != NULL ) {
		// free memory
		delete compute;
	}

	if( mover != NULL ) {
		// free memory
		delete mover;
	}
}

/**
 * Read input file with given name and create necessary
 */
bool Simulation::readInputFile(string inputFN) {
	inputFileName = inputFN;

	// read content of file into tree of block...
	inFile.setFileName(inputFN);
	inFile.read();

	//
	// ...now create all necessary object and fill them with data from input file
	//
	// is it test mode for input file
	inputFileTest = inFile.getInputFileTestMode();

	// create objects for species
	species = inFile.getSpecies("Species");

	if( species.size() == 0 ) {
		// no species defined
		cout << "There were no species defined in input file." << endl;
	}

	// create objects for region (plasma device)
	region = inFile.getSpatialRegion("SpatialRegion");

	bool regionDefined = region.defined(false);

	if( regionDefined ) {
		switch( region.get_compute_method() ) {
		default:
			// default value is green
			compute = new Compute_green();
			break;
		}
	}

	// define a mover (maybe there will be different movers in future)
	mover = new ParticleMover();

	// get values from spatial region
	maxSimStep = region.get_maxStep();
	dt = region.get_dt();

	for( unsigned int i=0 ; i < species.size() ; i++ ) {
		// according to p2c (ratio between real and computer particles), change species in system
		species[i].set_p2c_ratio( region.get_p2c() );
	}

	// get sources
	sources = inFile.getSources("Source", species);

	for( unsigned int i=0 ; i < sources.size() ; i++ ) {
		sources[i].set_simulation_dt(dt);
	}

	// get tree generation parameters
	tree = inFile.getTree("TreeCode");

	// store initial load (initial particles are created later -- in init)
	initLoad = inFile.getInitLoad("InitLoad", species);

	for( unsigned int i=0 ; i < initLoad.size() ; i++ ) {
		initLoad[i].set_dt(dt);
	}

	// read diagnostics data: what, when and how to output results/data
	diags = inFile.getDiagnostics("Diagnostics", species);

	for( unsigned int i=0 ; i < diags.size() ; i++ ) {
		// define some values to diagnostics sub-systems
		diags[i].set_max_step( maxSimStep );
		diags[i].set_dt( dt );
		diags[i].setSimulationDataAccess(this);

		// diags[i].dump(0);
	}

	if( inputFileTest == true ) {
		// display content of input file and exit
		for( unsigned int i=0 ; i < species.size() ; i++ ) {
			species[i].dump(0, i);
		}

		region.dump(0);

		for( unsigned int i=0 ; i < sources.size() ; i++ ) {
			sources[i].dump(0, i);
		}

		tree.dump(0);

		for( unsigned int i=0 ; i < initLoad.size() ; i++ ) {
			initLoad[i].dump( i );
		}

		for( unsigned int i=0 ; i < diags.size() ; i++ ) {
			diags[i].dump(0);
		}

		cout << endl << "End of input file printout. Exiting..." << endl;
		exit(0);
	}

	return 0;
}


/**
 * Initialize the simulation based on current data in simulation (particles, device, etc.).
 * Basically set simulation to point zero: reset time, inject initial particles, etc.
 */
void Simulation::init() {
	unsigned int i, j, max_i, max_j;
	ParticleList injParts, tmp_pl;

	cout << "Simulation :: init..." << endl;

	// start simulation with simulation step 0
	simStep = 0;
	t = 0.;
	injParts.init();

	for( i=0, max_i=sources.size() ; i < max_i ; i++ ) {
		// first initialize sources -- if we have some distributions (initialize parameters for distributions)
		sources[i].prepare();
	}

	for( i=0, max_i=initLoad.size() ; i < max_i ; i++ ) {
		// we have some definition of initial load; inject particles for initial load
		// inject particles from all definitions
		for( j=0, max_j=species.size() ; j < max_j ; j++ ) {
			tmp_pl = initLoad[i].inject( particles, tree, region, j, 0, 0 );
			// add injected particles to injected list
			injParts += tmp_pl;
		}
	}

	// now add injected particles to existing particles list
	particles += injParts;
	// delete temp particle lists
	injParts.init();
	tmp_pl.init();

	// particles.dump(0);

	// create tree for all particles in system
	tree.build( particles, region );

    // tree.dump(0);
	//dispMemoryUsage(true);

	// calculate forces and potential
	compute->forces( particles, tree, region );
	// move particles: calculate new positions and velocities
	mover->move( particles, region, t, dt, -1 );
	// compute potential
	potential = compute->potential( particles, tree, region );

#ifdef DEBUG
	dump.display( particles, species, "After particle injection" );
#endif

	// save to dump file if needed
//	diag.saveDump( .... sim. step, t );
	//dump.display( particles, species, "After first step" );
//	dump.display( potential, region );
//	dump.toFile("phase_space.dat", particles, species );
//	dump.toFile("potential.dat", potential, region );

	//dispMemoryUsage(true);

	// check (and write) some diagnostics
	writeDiagnostics();
}

/**
 * Run the simulation from current state (current particles and other data in system)
 */
int Simulation::run() {
	char comment[129];
	ParticleList injParts, tmp_pl;

	cout << "Simulation :: run" << endl;

    cout << "num. simulation steps: " << maxSimStep << endl;
    
	if( species.size() == 0 ) {
		// no species defined;
		cout << "No species or particles defined." << endl;
		return 1;
	}

	if( !region.defined(true) ) {
		// no region defined or not enough data
		cout << "Spatial region is not defined or there are data missing." << endl;
		return 1;
	}

	// setup progress indicator
	progress.setMaxStep(maxSimStep);

#ifdef DEGBUG
	// display dot for every step
	progress.addIndicator( 0, "." );
#else
	// setup progress: max step, dots for 5% mark, display percents every 20%.
	progress.addIndicator( 0.05, "." );
	progress.addIndicator( 0.2, "%.f%% " );
#endif
	//	progress.addIndicator( 0.05, "potential.dat", "potential" );

	// printf( "Max. number of computer particles: %g\n", particles.max_size() * 1. ) ;

	for( ; simStep <= maxSimStep ; simStep++ ) {
		// run simulation steps...
		// advance system simulation time for dt
		t += dt;
		injParts.init();

		// write current simulation step and time into status file... ltc_status.log
		for( unsigned int i=0, max_i=sources.size() ; i < max_i ; i++ ) {
			// inject particles in all sources, if needed
			if( simStep == 0 ) sources[i].dump(0,i);

			for( unsigned int j=0, max_j=species.size() ; j < max_j ; j++ ) {
				tmp_pl = sources[i].inject( particles, tree, region, j, t, simStep );
				// add injected particles to injected list
				injParts += tmp_pl;
			}
		}

		// now add injected particles to existing particles list
		particles += injParts;

		// particles.dump(0);

		// create tree for all particles in system
		tree.build( particles, region );

		// tree.dump(0);

		// calculate forces, positions and velocities
		compute->forces( particles, tree, region );
		// move particles: calculate new positions and velocities
		mover->move( particles, region, t, dt, simStep );

		// count number of particles outside right- & left- hand wall
		numParticlesLeftOut = region.countOutLeft( particles );
		numParticlesRightOut = region.countOutRight( particles );

		// check situation on boundaries (reflection and/or absorption of particles)
		region.checkBoundaries( particles );
		// compute potential
		potential = compute->potential( particles, tree, region );
		// clear generated tree
		tree.freeRoot();

		// save to dump file, if needed
		//	diag.saveDump( .... sim. step, t );
		//sprintf( comment, "After step: %ld", simStep);
		//dump.display( particles, species, comment );
		// display current progress
		progress.displayIndicator(simStep);

		if( simStep % 20 == 0 ) {
			// dispMemoryUsage(true);
		}

//		cout << simStep << " num. particles: " << particles[0].size() << " -- " << particles[0].capacity() << endl;

#ifdef DEBUG
		sprintf( comment, "step %ld", simStep );
		dump.display( particles, species, comment );
#endif
		// check (and write) some diagnostics
		writeDiagnostics();
	}

//	dump.display( particles, species, "End of simulation" );
//	dump.display( potential, region, "End of simulation" );
//	dump.toFile("phase_space.dat", particles, species );
//	dump.toFile("potential.dat", potential, region );

//	dump.display( particles, species, "End of simulation." );

	//dispMemoryUsage(true);

	return 0;
}


/**
 * Read data from dump file to enable user to continue the simulation state in dump file.
 * Also prepare data (similar to Init()) so simulation can continue with sim->Run()
 */
bool Simulation::readFromDumpFile(string inDumpFilename) {
	cout << "Simulation::readFromDumpFile( " << inDumpFilename << ") -- TO-DO!!" << endl;

	return true;
}


/**
 * Write diagnostics. Check all defined diagnostics and write data to file, if necessary.
 */
void Simulation::writeDiagnostics() {

	for( unsigned int i=0 ; i < diags.size() ; i++ ) {
		// check (and write) some diagnostics
		diags[i].save(simStep, t);
	}
}

/**
 * Display current memory usage.
 * @param detail -- detail usage (display each section separately) or only total
 */
void Simulation::dispMemoryUsage(bool details) {
	cout << getMemoryUsage(details);
}


/**
 * Get current memory usage and return it as string for displaying.
 * @param detail -- detail usage (display each section separately) or only total
 */
std::string Simulation::getMemoryUsage(bool details) {
	long long mem_fp = 0, total =  sizeof(this);
	unsigned int i;
	stringstream ret;

	if( details ) {
		ret << "Classes/structures: " << endl;
		ret << getMemUsageNumberFormated(" - Particle: ", sizeof(Particle));
		ret << getMemUsageNumberFormated(" - Vector3: ", sizeof(Vector3));
		ret << getMemUsageNumberFormated(" - ParticleList: ", sizeof(ParticleList));
		ret << getMemUsageNumberFormated(" - Species: ", sizeof(Species));
		ret << getMemUsageNumberFormated(" - Tree: ", sizeof(Tree));
		ret << getMemUsageNumberFormated(" - TreeNode: ", sizeof(TreeNode));
		ret << getMemUsageNumberFormated(" - Source: ", sizeof(Source));
		ret << endl;
	}

	ret << "Current data:" << endl;

	mem_fp = inFile.memory_footprint();
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - input file: ", mem_fp);

	mem_fp = progress.memory_footprint();
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - progress: ", mem_fp);

	mem_fp = sizeof(SpatialRegion);
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - spatial region: ", mem_fp);

	mem_fp = sizeof(ParticleMover);
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - particle mover: ", mem_fp);

	mem_fp = sizeof(Compute_green);
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - compute (green): ", mem_fp);

	mem_fp = sizeof(Dump);
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - dump: ", mem_fp);

	//mem_fp = tree.memory_footprint();
	//total += mem_fp;
	//if( details ) ret << getMemUsageNumberFormated(" - tree: ", mem_fp);

	mem_fp = particles.memory_footprint();
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - particles: ", mem_fp);

	mem_fp = species.size()* sizeof(Species);
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - species: ", mem_fp);

	for( mem_fp=0, i=0 ; i < sources.size() ; i++) {\
		mem_fp += sources[i].memory_footprint();
	}
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - source: ", mem_fp);

	mem_fp = inFile.memory_footprint();
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - input file: ", mem_fp);

	for( mem_fp=0, i=0 ; i < initLoad.size() ; i++) {\
		mem_fp += initLoad[i].memory_footprint();
	}
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - initial load: ", mem_fp);

	mem_fp = potential.size()*sizeof(double);
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - potential: ", mem_fp);

	mem_fp = fields.size()*sizeof(Field);
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - fields: ", mem_fp);

	for( mem_fp=0, i=0 ; i < diags.size() ; i++) {\
		mem_fp += diags[i].memory_footprint();
	}
	total += mem_fp;
	if( details ) ret << getMemUsageNumberFormated(" - diagnostics: ", mem_fp);

	ret << getMemUsageNumberFormated("Total: ", total);

	return ret.str();
}


/**
 * Display memory usage in bytes. In number if high enough also in kilobytes, megabytes and gigabytes.
 */
std::string Simulation::getMemUsageNumberFormated(string text, long long total) {
	stringstream ret;
	double total_k, total_m, total_g;

	total_k = total / 1024;
	total_m = total_k / 1024;
	total_g = total_m / 1024;

	ret << text << total << " b";
	if( total_k > 1. ) ret << " (" << total_k << " kb";
	if( total_m > 1. ) ret << ", " << total_m << " Mb";
	if( total_g > 1. ) ret << ", " << total_g << " Gb";
	if( total_k > 1. ) ret << ")";

	ret << endl;
	return ret.str();
}

