/*
 * Source.cpp
 *
 *  Created on: Aug 7, 2011
 *      Author: janez
 */

#include <string>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "Source.h"
#include "ParticleList.h"
#include "RandomGenerator.h"

#define	DISTRIBUTION_CELLS		8192

using namespace std;


Source::Source() {
	init();
}

Source::~Source() {

}

/**
 * Initialize source object.
 */
void Source::init() {
	type = TYPE_DELTA_UNDEFINED;
	position_dist = POSITION_DISTRIBUTION_UNIFORM;
	velocity_dist = VELOCITY_DISTRIBUTION_NONE;
	numPartToInject = inject_dStep = 0;
	inject_dt = simulation_dt = 0.;
	lastInjectTime = 0;
	lastInjectStep = 0;
	inject_r.set(0.,0.,0.);
	inject_v.set(0.,0.,0.);
	inject_T.set(0.,0.,0.);
	sp.init();
	sp_index = -1;
	v_distribution.clear();
}

/**
 * Memory occupied by current object.
 */
long long Source::memory_footprint() {
	return sizeof(Source) + v_distribution.size()*sizeof(double);
};


/**
 * Check if all necessary data for source/initLoad are defined.
 */
std::vector<std::string> Source::definedDataOK(std::vector<Species> &species) {
	std::vector<std::string> errors;
	int sp_i;


	// make return vecto empty
	errors.clear();

	if( numPartToInject == 0 ) {
		errors.push_back( "number of particles to inject is not defined; to disable source, enter 'disabled' inside block, not number of particles to 0");
	}

	if( type == TYPE_DELTA_UNDEFINED ) {
		// injection type (when to do injection) is not defined
		errors.push_back( "injection step not defined (define 'dt' or 'di')" );
	}

	if( sp_index == -1 ) {
		// species not defined
		errors.push_back( "species not defined (define with 'species = <species name>')" );
	} else {
		// check if given species is define in lost of species
		if( (sp_i = isSpeciesDefined( sp.get_name(), species )) == -1 ) {
			// given species does not exist;
			errors.push_back( "given species '" + sp.get_name() + "' is not define (check species name of species definition at beginning of file)" );
		}
	}

	return errors;
}


/**
 * Check if species with given name is defined and return index of species.
 * If species is not defined, return -1
 * @return 0 or more: index of species in list of species, -1: not defined
 */
int Source::isSpeciesDefined(std::string name, std::vector<Species> &species) {
	int pos = -1;
	unsigned i;


	for( i=0 ; i < species.size() && pos == -1; i++ ) {
		/* check all defined species */
		if( species[i].get_name() == name ) {
			/* OK, species found; return position */
			pos = i;
		}
	}

	return pos;
}


/**
 * Prepare source distribution.
 */
void Source::prepare() {
	double v_off;

	//
	// initialize velocity distributions
	//
	switch( velocity_dist ) {
	case VELOCITY_DISTRIBUTION_MAXWELLIAN:
		// prepare distribution factors and velocities for Maxwellian velocity distribution
		if( ! IS_ZERO(inject_T.e1()) ) {
			// temperatures defined; compute necessary things...
			double vt_inj = sqrt(fabs(inject_T.e1()*sp.get_q()/sp.get_m()));
			/* shift velocity */
			v_off = inject_v.e1() / vt_inj;
			/* create distribution */
			prepareMaxwellianDistributionFactors(0., 5.+v_off, DISTRIBUTION_CELLS, v_off, false);

			for( unsigned int i = 0; i < v_distribution.size(); i++ ) {
				// compute velocities...
				v_distribution[i] *=  vt_inj;
			}
		}

		break;

	}



}



/**
 * Print content of source in human readable format.
 */
void Source::dump(int level, int index) {
	string offset = "";
	string disp_type[] = { "", "none", "time", "step" };


	for( int i=0 ; i < level ; i++ ) {
		offset += "  ";
	}

	cout << endl;
	cout << offset << "Source:" << endl;
	cout << offset << "=======" << endl;
	cout << offset << "Block index in input file: " << index << endl;
	cout << offset << "Species index: " << sp_index << endl;
	cout << offset << "Number of particles to inject in single injection: " << numPartToInject << endl;
	cout << offset << "Injection criteria: " << disp_type[type] << endl;

	if( type == TYPE_DELTA_TIME ) {
		cout << offset << "  Time between injections: " << inject_dt << endl;
	} else if( type == TYPE_DELTA_STEP ) {
		cout << offset << "  Steps between injections: " << inject_dStep << endl;
	}

	// dump position and velocity distributions and other data
	dumpPositions(offset);
	dumpVelocities(offset);
}

/**
 * Print content of source in human readable format.
 */
void Source::dumpPositions(string offset) {
	string disp_position[] = { "none (constant)", "uniform", "random", "uniform, no borders", "uniform, no left border", "uniform, no right border" };

	cout << offset << "Position distribution: ";

	if( position_dist < POSITION_DISTRIBUTION_COUNT ) {
		cout << disp_position[position_dist] << endl;
	} else {
		cout << "UNKNOWN DISTRIBUTION!" << endl;
		return;
	}

	switch( position_dist ) {
	case POSITION_DISTRIBUTION_NONE:
		// single position
		cout << offset << "  injecting position (x, y, z): ";
		cout << inject_r.e1() << ", " << inject_r.e2() << ", " << inject_r.e3() << endl;
		break;

	case POSITION_DISTRIBUTION_UNIFORM:
		// range of positions
		cout << offset << "  injecting particles uniformly between border positions, including borders (edge positions)" << endl;
		cout << offset << "  start position (x, y, z): ";
		cout << inject_r_min.e1() << ", " << inject_r_min.e2() << ", " << inject_r_min.e3() << endl;
		cout << offset << "  end position (x, y, z): ";
		cout << inject_r_max.e1() << ", " << inject_r_max.e2() << ", " << inject_r_max.e3() << endl;
		break;

	case POSITION_DISTRIBUTION_RANDOM:
		// range of positions
		cout << offset << "  injecting particles at random positions between given limits" << endl;
		cout << offset << "  start position (x, y, z): ";
		cout << inject_r_min.e1() << ", " << inject_r_min.e2() << ", " << inject_r_min.e3() << endl;
		cout << offset << "  end position (x, y, z): ";
		cout << inject_r_max.e1() << ", " << inject_r_max.e2() << ", " << inject_r_max.e3() << endl;
		break;

	case POSITION_DISTRIBUTION_UNIFORM_NO_BORDERS:
		// range of positions
		cout << offset << "  injecting particles uniformly between positions - without borders (edge positions)" << endl;
		cout << offset << "  start position (x, y, z): ";
		cout << inject_r_min.e1() << ", " << inject_r_min.e2() << ", " << inject_r_min.e3() << endl;
		cout << offset << "  end position (x, y, z): ";
		cout << inject_r_max.e1() << ", " << inject_r_max.e2() << ", " << inject_r_max.e3() << endl;
		break;

	case POSITION_DISTRIBUTION_UNIFORM_NO_BORDER_LEFT:
		// range of positions
		cout << offset << "  injecting particles uniformly between positions - without left border, including right one" << endl;
		cout << offset << "  start position (x, y, z): ";
		cout << inject_r_min.e1() << ", " << inject_r_min.e2() << ", " << inject_r_min.e3() << endl;
		cout << offset << "  end position (x, y, z): ";
		cout << inject_r_max.e1() << ", " << inject_r_max.e2() << ", " << inject_r_max.e3() << endl;
		break;

	case POSITION_DISTRIBUTION_UNIFORM_NO_BORDER_RIGHT:
		// range of positions
		cout << offset << "  injecting particles uniformly between positions - including left border but without right one" << endl;
		cout << offset << "  start position (x, y, z): ";
		cout << inject_r_min.e1() << ", " << inject_r_min.e2() << ", " << inject_r_min.e3() << endl;
		cout << offset << "  end position (x, y, z): ";
		cout << inject_r_max.e1() << ", " << inject_r_max.e2() << ", " << inject_r_max.e3() << endl;
		break;

	}

}

/**
 * Print content of source in human readable format.
 */
void Source::dumpVelocities(string offset) {
	string disp_velocity[] = { "none (constant)", "maxwellian", "random", "random, cold" };

	cout << offset << "Velocity distribution: ";

	if( velocity_dist < VELOCITY_DISTRIBUTION_COUNT ) {
		cout << disp_velocity[velocity_dist] << endl;
	} else {
		cout << "UNKNOWN DISTRIBUTION!" << endl;
		return;
	}

	switch( velocity_dist ) {
	case VELOCITY_DISTRIBUTION_NONE:
		// no distribution (single velocity)
		cout << offset << "  Injection velocity (vx, vy, vz): ";
		cout << inject_v.e1() << ", " << inject_v.e2() << ", " << inject_v.e3() << endl;
		break;

	case VELOCITY_DISTRIBUTION_MAXWELLIAN:
		// display middle velocity + temperatures
		cout << offset << "  offset injection velocity (vx, vy, vz): ";
		cout << inject_v.e1() << ", " << inject_v.e2() << ", " << inject_v.e3() << endl;
		cout << offset << "  temperatures [eV]: " << endl;
		cout << offset << "    - parallel (Tx): " << inject_T.e1() << endl;
		cout << offset << "    - perpendicular (Tz): " << inject_T.e2() << endl;
		break;

	case VELOCITY_DISTRIBUTION_RANDOM:
		// range of velocities
		cout << offset << "  Injecting velocity between velocities, including given ones" << endl;
		cout << offset << "  start velocity (vx, vy, vz): ";
		cout << inject_v_min.e1() << ", " << inject_v_min.e2() << ", " << inject_v_min.e3() << endl;
		cout << offset << "  end velocity (vx, vy, vz): ";
		cout << inject_v_max.e1() << ", " << inject_v_max.e2() << ", " << inject_v_max.e3() << endl;
		break;

	case VELOCITY_DISTRIBUTION_RANDOM_COLD:
		// random velocities + direction
		cout << offset << "  Injecting velocity from 0 to given value: random size and direction" << endl;
		cout << offset << "  max velocity (vx, vy, vz): ";
		cout << inject_v.e1() << ", " << inject_v.e2() << ", " << inject_v.e3() << endl;
		break;
	}
}


/**
 * Inject particles if it is needed.
 * plist -- particle list
 * in_region -- region in which to inject particle(s)
 * in_index -- species index that we have to inject
 * t -- current simulation time
 * step -- current simulation time
 */
ParticleList Source::inject( ParticleList& plist, Tree &in_tree, SpatialRegion& in_region, int in_sp_index, double t, long int step ) {
	bool inject = false;
	ParticleList retList;

	// initialize return list
	retList.init();

	if( in_sp_index != this->sp_index ) {
		// OK, current source does not add requested species
		return retList;
	}

	switch( type ) {
	case TYPE_DELTA_TIME:
		// check for injecting based on time
		if( t >= lastInjectTime + inject_dt ) {
			// yes, we have to inject this time
			lastInjectTime = t;
			inject = true;
		}

		break;

	case TYPE_DELTA_STEP:
		// check for injecting based on current step
		if( step >= lastInjectStep + inject_dStep ) {
			// yes, we have to inject this time
			lastInjectStep = step;
			inject = true;
		}

		break;

	default:
		// unknown injection step; do nothing
		inject = false;
		break;
	}

	if( inject ) {
		// inject "numPartToInject" particles in system
		retList = doInjectParticles( plist, in_tree, in_region, t, step);
	}

	return retList;
}


/**
 * Inject particles with current data (number of particles and for defined species).
 * Method does not check if injections should be done - this is done before.
 * plist -- particle list
 * in_region -- region in which to inject particle(s)
 * in_index -- species index that we have to inject
 * t -- current simulation time
 * step -- current simulation time
 */
ParticleList Source::doInjectParticles( ParticleList& plist, Tree &in_tree, SpatialRegion& in_region, double t, long int step ) {
	Vector3 v, r, v_comp, r_comp, F;
	Particle new_p;
	double inj_dt;
	Compute_green compute;
	ParticleList retList;

	// initialize return list
	retList.init();

	inj_dt = simulation_dt / numPartToInject;

	for( long int i=0 ; i < numPartToInject ; i++ ) {
		// get initial position and velocity for new particle
		// define values for new particle
		new_p.init();
		new_p.set_q( sp.get_q() );
		new_p.set_m( sp.get_m() );
		new_p.set_sp_index( this->sp_index );

		// compute velocity and position according to distribution
		v_comp = inject_velocity(i);
		r_comp = inject_position(i);

		if( i == 0 ) {
			// make correction to velocity for first particle (according to LeapFrog method)
			F = compute.force(plist, in_region, new_p);
			new_p.set_f( F );
			v = v_comp + (-simulation_dt/2.) * new_p.get_f() / new_p.get_m();

			F = compute.force(plist, in_region, new_p);
			new_p.set_f( F );
			r = r_comp + inj_dt/2. * v;
			v = v_comp + (inj_dt/2. * new_p.get_f() ) / new_p.get_m();
		} else {
			// other particles; no shift of velocity
			F = compute.force(plist, in_region, new_p);
			new_p.set_f( F );
			r = r_comp + inj_dt * v_comp;
			v = v_comp + (inj_dt * new_p.get_f() ) / new_p.get_m();
		}

		// define values for new particle
		new_p.set_r(r);
		new_p.set_v(v);
		// add particle to particle list
		retList += new_p ;
	}

	return retList;
}




/**
 * Calculate position for new particle.
 */
Vector3 Source::inject_position(int pi) {
	Vector3 r, dr;


	if( pi < 0 || numPartToInject <= pi ) {
		// given index is wrong; display error and return 0 for position
		cout << "Error: inject_position - particle index " << pi << " is larger then number of particles to inject " << numPartToInject << "." << endl;
		r.set(0., 0., 0.);
		return r;
	}

	switch( position_dist ) {
	case POSITION_DISTRIBUTION_RANDOM: {
		// random positions between r_min and r_max;
		// define random generator and initialize it (define a seed)
		RandomGenerator rnum;

		dr = inject_r_max - inject_r_min;
		// get random factor and compute new position
		r = inject_r_min + rnum.generate() * dr;

		break;
	}

	case POSITION_DISTRIBUTION_UNIFORM:
		// uniform distribution between r_min and r_max;
		// one particle is on left border and one is on right border;
		if( numPartToInject == 1 ) {
			// if we have to inject only one particle, it is positioned in the middle of range
			for( int i=0 ; i < 3 ; i++ ) {
				// compute position for all coordinates
				r.set( i+1, 0.5 * (inject_r_min.e(i+1) + inject_r_max.e(i+1)) );
			}

		} else {
			// compute particle position
			// compute difference between particles
			dr = inject_r_max - inject_r_min;
			// here we have n-1 - this assures that first and last particle is on border of range
			dr /= (numPartToInject-1);
			// compute new position
			r = inject_r_min + pi * dr;
		}

		break;

	case POSITION_DISTRIBUTION_UNIFORM_NO_BORDERS:
		// uniform distribution between r_min and r_max;
		// there are no particles on borders - only inside range
		dr = inject_r_max - inject_r_min;
		// here we have n+1 - this assures that no particle is on border of range
		dr /= (numPartToInject+1);
		// compute new position
		r = inject_r_min + (pi + 1) * dr;

		break;

	case POSITION_DISTRIBUTION_UNIFORM_NO_BORDER_LEFT:
		// uniform distribution between r_min and r_max;
		// there is no particle on LEFT border of range
		dr = inject_r_max - inject_r_min;
		// here we have n- this assures that there is no particle on one border
		dr /= (numPartToInject);
		// compute new position; there is i+1 -> first point starts inside range
		r = inject_r_min + (pi + 1) * dr;

		break;

	case POSITION_DISTRIBUTION_UNIFORM_NO_BORDER_RIGHT:
		// uniform distribution between r_min and r_max;
		// there is no particle on RIGHT border of range
		dr = inject_r_max - inject_r_min;
		// here we have n- this assures that there is no particle on one border
		dr /= (numPartToInject);
		// compute new position; there is i -> first point starts on border
		r = inject_r_min + pi * dr;

		break;

	case POSITION_DISTRIBUTION_NONE:
	default:
		// unknown type for positions
		// single position for all particles
		r = inject_r;
		break;
	}

	return r;
}


/**
 * Calculate velocity for new particle.
 */
Vector3 Source::inject_velocity(int pi) {
	Vector3 v, dv;
	// define and initialize random generator
	RandomGenerator rnum;

	switch( velocity_dist ) {
	case VELOCITY_DISTRIBUTION_RANDOM: {
		// random velocities between v_min and v_max
		dv = inject_v_max - inject_v_min;
		// get random factor and compute new position
		v = inject_v_min + rnum.generate() * dv;

		break;
	}

	case VELOCITY_DISTRIBUTION_MAXWELLIAN: {
		// velocities according Maxwellian
		// define also other velocity components...
		double v_perp, theta, vt_perp;

		if( ! IS_ZERO(inject_T.e1()) ) {
			// temperature in x-direction defined; "hot" beam
			int pos = DISTRIBUTION_CELLS * rnum.generate();
			// define "x" component of velocity
			v.set_e1( v_distribution[pos] );

		} else {
			// no temperature; "cold" beam
			v.set_e1( inject_v.e1() );
		}

		// compute velocity based on particle temperature (hot beam)
		vt_perp = sqrt(fabs(inject_T.e3()*sp.get_q()/sp.get_m()));
		v_perp = vt_perp * sqrt(2.*fabs(log(rnum.generate()+1e-15)));
		theta = 2 * PI * rnum.generate();
		v.set_e2( v_perp * sin(theta) );
		v.set_e3( v_perp * cos(theta) + inject_v.e3() );
		break;
	}

	case VELOCITY_DISTRIBUTION_RANDOM_COLD: {
		// velocities random with random angles...
		// define also other velocity components...
		double r, theta;

		do {
			r = sqrt(-2.*log(rnum.generate()+1.e-20));
		} while( r > 5. );

		theta = 2 * PI * rnum.generate();
		v.set_e2( r * cos(theta) * inject_v.e1() );
		v.set_e3( r * sin(theta) * inject_v.e1() );

		do {
			r = sqrt(-2.*log(rnum.generate()+1.e-20));
		} while( r > 5. );

		theta = 2 * PI * rnum.generate();
		v.set_e1( r * cos(theta) * inject_v.e1() );
		break;
	}

	case VELOCITY_DISTRIBUTION_NONE:
	default:
		// single velocity for all particles
		v = inject_v;
		break;
	}

	return v;
}



/**
 * Compute velocity distribution for Maxwellian distribution.
 * Function taken from BIT1 and adopted for use here.
 */
void Source::prepareMaxwellianDistributionFactors(double vmin, double vmax, int size, double v_off, bool force){
	int    ii;
	float  v1;
	double vv, ddv, sf, ff, ddf, sumf, ss, ds;


	if( v_distribution.size() == 0 || (v_distribution.size() > 0 && force) ) {
		// we have to compute distribution coefficients...
		if( v_distribution.size() > 0 ) {
			v_distribution.clear();
		}

		// allocate requested number of cells
		v_distribution.reserve( size );

		/*
		 * J.K. 2012-05-16 -
		 * I hope this is correct:
		 *  1. compute area below given function (= sumf)
		 *  2. explode computed area into "size" parts and store velocities at which there is border between cells
		 */
		v1 = vmin;
		ddv = 0.5*(vmax-v1)/size;
		vv = v1;
		sf = 0.5*fvmaxw(vv, v_off);

		for (ii=0; ii<size; ii++) {
			vv += ddv;
			ddf = fvmaxw(vv, v_off);
			sf += ddf + ddf;
			vv += ddv;
			ddf = fvmaxw(vv, v_off);
			sf += ddf;
		}

		ddf = 0.5*fvmaxw(vmax, v_off);
		sf -= ddf;
		sf = (sf + sf)*ddv/3.;             /* Cumulative function */

		sumf = 0.;
		ii = 0;
		vv = v1;
		ds = sf/size;
		ss = ds/2.;

		while( ii < size ) {
			// fill all requested cells
			ff = fvmaxw(vv, v_off);
			ddv = 1./128.;

			if (ff > 0.) {
				ddv = min(ddv, .45*ds/ff);
			}

			do {
				ddf = fvmaxw(vv, v_off) + 4.*fvmaxw(vv+ddv, v_off) + fvmaxw(vv+ddv+ddv, v_off);
				ddf *= ddv/3.;
				if (ddf >= ds) ddv *= 0.5;
			} while (ddf >= ds);


			vv += ddv+ddv;
			sumf += ddf;

			if (sumf < ss) {
				// loop
			} else {
				/* When a desired velocity has been passed, interpolate
				 * to find it, and start looking for the next velocity */
				v_distribution.push_back(vv - (sumf - ss)*ddv/ddf);
				ii++;
				ss += ds;
			}
		}
	}

}

