/**
 * Tree.cpp
 *
 *  Created on: Aug 22, 2011
 *      Author: janez
 */

#include <string>
#include <iostream>
#include <stdlib.h>
#include <cstdio>

#include "Tree.h"


using namespace std;

//#define DEBUG_ALLOC

Tree::Tree() {
	init();

}

Tree::~Tree() {
	// free memory occupied by tree nodes
	delete root;
}

/**
 * Initialize object
 */
void Tree::init() {
	maxParticlesInNode = 0;
	node2SystemLengthRatio = 1.;
	p2cRatio = 0.;
	root = NULL;
	numNodes = maxLevel = 0;
	treecode_disabled = false;
}

/**
 * Dump tree info.
 */
void Tree::dump(int level) {
	string offset = "";

	for( int i=0 ; i < level ; i++ ) {
		offset += "  ";
	}

	cout << endl << offset << "Tree parameters: " << endl;
	cout << offset << "  For computing use: ";

	if( treecode_disabled == false ) {
		cout << "treecode" << endl;
	} else {
		cout << "direct summation" << endl;
	}

	cout << offset << "  Max. particles in node: " << maxParticlesInNode << endl;
	cout << offset << "  node length ratio: " << node2SystemLengthRatio << endl;
	cout << offset << "  p2c ratio: " << p2cRatio << endl;
	cout << offset << "  root node defined: ";

	if( this->root != NULL ) {
		cout << "yes" << endl;
	} else {
		cout << "no" << endl;
	}

	cout << "  Max. level: " << maxLevel << ", total number of nodes: " << numNodes << endl;
	cout << "  Number of particles in tree: " << this->count_particles() << endl;

	if( this->root != NULL ) {
		this->root->dump( 0 );
	}
}


/**
 * Build tree for given list of particles.
 */
void Tree::build( ParticleList& plist, SpatialRegion& region ) {
	Grid grid = region.get_grid();
	Vector3 r0, r1;
	std::vector<int> dummy;

	maxLevel = numNodes = 0;
	r0 = grid.get_r0();
	r1 = grid.get_r1();
	dummy.clear();

	root = buildTree( plist, region, r0.e1(), r1.e1(), dummy, 0, &maxLevel, &numNodes );

	// calculate values in all nodes (Qc, Xc)
//	computeQc( plist, root );
//	computeXc( root );
}



/*
 * Build node for tree.
 * Store indexes of particles that are inside given limits (max_x and min_x), compute total charge of node (Qc) and
 * center point of node (Xc). Last two number are used in computing force and potential.
 *
 */
TreeNode* Tree::buildTree( ParticleList& parts, SpatialRegion& region, double ds, double de, std::vector<int>& parent_parts, long level, long *maxLevel, long *numNodes ) {
	TreeNode *node = NULL;	/* initialize as empty node */
	std::vector<int> particles_inside;
	int indx;
	unsigned int i, max_i;
	double node_ds, node_de, midp, sumXc, Qc, tmp_x;


	if( parts.size() == 0 ) {
		// no particles
		return NULL;
	}

	if( parent_parts.size() == 0 ) {
		/* we did not get any particles from parent node; it seams we are at root node; use all particles (from parts) */
#ifdef DEBUG_ALLOC
	printf("[in buildTree] from parent; level: %ld, number of particles: %ld; node dim: %f --> %f\n",
			level, parts.size(), ds, de );
#endif
		/* initialize few counters */
		*maxLevel = 0;
		*numNodes = 0;

		/* allocate space for temporary list of indexes of particles, that are inside given domain */
		particles_inside.reserve(parts.size());

		for( i=0, Qc=0, sumXc=0, max_i=parts.size() ; i < max_i ; i++ ) {
			/* check which particles are inside given domain */
			tmp_x = parts.getParticle(i)->get_r().e1();

//printf("%d: %lf -- %lf : x=%lf\n", i, ds, de, pt_part->x);

			if( ds <= tmp_x && ( tmp_x < de ||  IS_ZERO(tmp_x - de) ) ) {
				/* test can be simple, because we are using all particles and limits for current node is whole system */
				/* ok, current particle is in, store it's index into temp array
				 * check for part.x < upper limit or part.x == upper limit, if testing at the end of system */
				particles_inside.push_back(i);
				/* calculate total charge and center point of cluster */
				Qc += parts.getParticle(i)->get_q();
				sumXc += tmp_x;
			}
		}

#ifdef DEBUG_ALLOC
		printf("[in buildTree] from parent; particles inside: %ld\n", particles_inside.size());
#endif

	} else {
		/* we are dividing some node; use only particles that are already defined at parent node */
#ifdef DEBUG_ALLOC
	printf("[in buildTree] subdividing node; level: %ld, number of particles: %ld; node dim: %f --> %f\n",
			level, parent_parts.size(), ds, de );
#endif
		/* allocate space for temporary list of indexes of particles, which are inside given domain */
		particles_inside.reserve(parent_parts.size());

		for( i=0, Qc=0, sumXc=0; i < parent_parts.size() ; i++ ) {
			/* check which particles are inside given domain */
			indx = parent_parts[i];
			tmp_x = parts.getParticle(indx)->get_r().e1();

#ifdef DEBUG_ALLOC
		printf("[in buildTree] subdividing node; x=%g, ds=%g, de=%g ", tmp_x, ds, de);
#endif

			if( ds <= tmp_x && ( tmp_x < de || (IS_ZERO(de - region.get_grid_x_max()) && IS_ZERO(tmp_x - de) ) ) ) {
				/* ok, current particle is in, store it's index into temp array */
				particles_inside.push_back(indx);
				/* calculate total charge and center point of cluster */
				Qc += parts.getParticle(indx)->get_q();
				sumXc += tmp_x;
			}
		}
#ifdef DEBUG_ALLOC
		printf("[in buildTree] subdividing node; particles inside: %ld\n", particles_inside.size());
#endif
	}

	/* initialize node with n_in number of particle indexes inside and add given particle indexes */
	node = new TreeNode(particles_inside);

#if 1
	/* define boundaries for particles in current node */
	shrinkBox( parts, node, &node_ds, &node_de );
#else
	node_ds = ds;
	node_de = de;
#endif
	node->setDs(ds);
	node->setDe(de);
	/* calculate total charge and center point of a node */
	node->setQc(Qc);

	if( node->getNumberOfParticles() > 0 ) {
		// store center of particles in node
		node->setXc( sumXc / node->getNumberOfParticles() );
	} else {
		// store middle point of node as center
		node->setXc( 0.5 * (node_ds + node_de) );
	}

	/* we created one node */
	(*numNodes) ++;
	node->setLevel(level);
	node->setIndex(*numNodes);

#ifdef DEBUG_ALLOC
//	printf("node no: %ld | particles: %d, limit: %d ; node dim: %f --> %f, node len: %f, system len: %f --> ratio: %f, limit: %f\n",
//			node->index, node->part_indx.size(), maxParticlesInNode, node->ds, node->de, node->de - node->ds, region.get_grid_x_max() - region.get_grid_x_min(),
//			fabs(node->de - node->ds)/fabs(region.get_grid_x_max() - region.get_grid_x_min()), node2SystemLengthRatio
//		);
#endif
	double nodeSize = fabs(node->getDe() - node->getDs());
	double systemSize = fabs(region.get_grid_x_max() - region.get_grid_x_min());

	if( !this->treecode_disabled && node->getNumberOfParticles() > this->maxParticlesInNode &&  nodeSize > systemSize * this->node2SystemLengthRatio ) {
		// if treecode is not disabled, go further; otherwise we only have one node with all particles in it.
		// node has too many particles or is too big in comparison to system length;
		// divide current node into sub-nodes
		if (level+1 > *maxLevel) {
			*maxLevel = level+1;
		}

#ifdef DEBUG_ALLOC
		printf(" --> node subdivide\n");
#endif
		TreeNode *tmp_node;
		/* we have to divide current node into two more */
		midp = 0.5*(node_ds + node_de);
		/* generate leafs for current node */
		tmp_node = buildTree(parts, region, node_ds, midp, particles_inside, level+1, maxLevel, numNodes);
		node->setLeaf1( tmp_node );
		tmp_node = buildTree(parts, region, midp, node_de, particles_inside, level+1, maxLevel, numNodes);
		node->setLeaf2( tmp_node );

		if( level > 0 ) {
			/* if level is more that one (not root node), delete particle list in node (we do not need it); only last nodes have
			 * particle list in it (we need it for calculations) */
			node->clearParticleIndexes();
		}
	}


#ifdef DEBUG_ALLOC
	printf("Number of particles: %ld, start: %lf, end: %lf\n", parts.size(), ds, de );
	printf("[in buildTree] TC tree created. Number of levels: %ld, total nodes: %ld\n", *maxLevel, *numNodes );
//	printf("indexes:");

//	for( i=0 ; i < node->part_indx.size() ; i++) {
		/* use all particles in current tree node */
//		printf(" %d", node->part_indx[i] );
//	}
//	printf("\n");
#endif
	// clear vector and free memory
	particles_inside.clear();

	return node;
}


/*
 * define smaller box (beginning and end of section) for given tree node
 */
void Tree::shrinkBox(ParticleList& parts, TreeNode *tree, double *ds, double *de) {
	unsigned int i;
	double tmp_ds, tmp_de;
	Vector3 r;

	if( tree->getNumberOfParticles() > 0 ) {
		/* we have some points in current node */
		r = parts.getParticle(tree->getParticleIndex(0))->get_r();
		tmp_ds = tmp_de = r.e1();

		for( i=0 ; i < tree->getNumberOfParticles() ; i++) {
			/* compare all points in given node */
			r = parts.getParticle(tree->getParticleIndex(i))->get_r();
			tmp_ds = MIN( tmp_ds, r.e1() );
			tmp_de = MAX( tmp_de, r.e1() );
		}

	} else {
		tmp_ds = tmp_de = 0;
	}

	/* store values before returning... */
	*ds = tmp_ds;
	*de = tmp_de;
	return;
}

/**
 * How many memory is tree currently using
 */
long long Tree::memory_footprint() {
	return sizeof(Tree) + this->root->memory_footprint();
};

/*
 * Count particles in tree - returns number of particles in tree (for checking).
 */
long Tree::count_particles() {

	if( this->root != NULL ) {
		// return number of particles
		return this->root->getNumberOfParticles();
	}

	return 0;
}
